/*
 * DataInfo.C	.MEX file
 * check information from a pre-existing M-file 
 * input: filename 
 * e.g. DataInfo('myfile.m')
 */

#include <stdio.h>
#include "mex.h"
#include "mexversion.h"

#define CHANNUM     16

void mexFunction(
                 int nlhs,       mxArray *plhs[],
                 int nrhs, const mxArray *prhs[]
		 )
{
int   j, i, Nchan, datatype, tempint;
int   yy, mm, dd, hh, mn, ss;
float tempf;
char  chName[CHANNUM][80], filename[100], tempst[100];
FILE *fp;

  /* Check for proper number of arguments */
  
  if (nrhs < 1 )
    mexErrMsgTxt("you need to input a data file name.");

  if ( !mxIsChar(prhs[0]) ) {
    mexErrMsgTxt("The first argument should be the filename");
  }
  mxGetString (prhs[0], filename, 100);

  /* open the file for information */

  fp = fopen(filename, "r+b");
  if ( fp == NULL ) {
    printf ( "Cann't open reading file: %s\n", filename );
    mexErrMsgTxt(" ");
  }
  else {
    printf ( "You're runing Data Mex version %3.1f\n", VERSION );
    printf ( "File '%s':\n", filename );
  }
  fscanf ( fp, "%s", tempst ); 
  if (strncmp(tempst, "%dataVersion", 11))
    mexErrMsgTxt("Not a file generated by Data Mex.");
  sscanf ( tempst, "%%dataVersion%f-%d", &tempf, &tempint );
  printf ( "    Generated by Data Mex version %3.1f\n", tempf );
  if ( tempint )
    printf ( "    binary file\n" );
  else
    printf ( "    text file\n" );

  if ( tempf - VERSION > 0.01 || VERSION - tempf > 0.01)
    printf ( "Warning: version numbers don't match!\n" );
  fscanf ( fp, "%s", tempst ); 
  sscanf ( tempst, "dataType=%d", &datatype );
  if ( datatype == 0 )
    printf ( "    Data Type raw data\n" );
  else if ( datatype == 1 )
    printf ( "    Data Type second trend data\n" );
  else 
    printf ( "    Data Type minute trend data\n" );
  fscanf ( fp, "%s", tempst ); 
  sscanf ( tempst, "dataChnum=%d", &Nchan );
  fscanf ( fp, "%s", tempst );
  sscanf ( tempst, "dataDuration=%d", &tempint );
  if (datatype == 60) 
    printf ( "    Duration %d minutes\n", tempint );
  else
    printf ( "    Duration %d seconds\n", tempint );
  fscanf ( fp, "%s", tempst );
  if (datatype == 0) {
    sscanf ( tempst, "dataResolution=%d", &tempint );
    printf ( "    Sampling rate %d\n", tempint );
  }
  else {
    sscanf ( tempst, "dataTrend=%d", &tempint );
    if ( tempint == 0 )
      printf ( "    Trend min mean max\n" );
    else
      printf ( "    Trend %s\n", trendText[tempint] );
  }
  fscanf ( fp, "%s", tempst );
  if (datatype == 0) {
    sscanf ( tempst, "dataFilter=%d", &tempint );
    printf ( "    Data Filter %d: ", tempint );
    switch ( tempint ) {
    case 0: 
        printf ( "no filter\n" );
        break;
    case 1: 
        printf ( "least-squares  42  0.9  0.015-0.2 dB\n" );
        break;
    case 2: 
        printf ( "equiripple  42  0.9  0.06 dB\n" );
        break;
    case 3: 
        printf ( "least-squares  22  0.9  0.1-1 dB\n" );
        break;
    case 4: 
        printf ( "least-squares  82  0.9  0.0006-.01 dB\n" );
        break;
    }  
  }
  else {
    sscanf ( tempst, "dataSkip=%d", &tempint );
    printf ( "    Skip total %d second\n", tempint );
  }

  printf ( "    Total %d channels\n", Nchan );
  fscanf ( fp, "%s", tempst ); 
  for ( j=0; j<Nchan-1; j++ ) {
    fscanf ( fp, "%s", chName[j] );
  } 
  fscanf ( fp, "%s", chName[Nchan-1] );
  printf ( "    %s", chName[0] );
  if ( Nchan>1 )
    for ( j=1; j<Nchan; j++ ) {
      printf ( "  %s", chName[j] );
    }
  printf ( "\n" );
  fscanf ( fp, "%s", tempst );
  fscanf ( fp, "%s", tempst );
  sscanf ( tempst, "dataStart='%d-%d-%d-%d-%d-%d';", &yy,&mm,&dd,&hh,&mn,&ss );
  if ( yy < 10 )
    sprintf ( tempst, "0%d-%d-%d-%d-%d-%d", yy,mm,dd,hh,mn,ss );
  else
    sprintf ( tempst, "%d-%d-%d-%d-%d-%d", yy,mm,dd,hh,mn,ss );
  printf ( "    Starting time %s\n", tempst );
  fclose(fp);

		 
  return;
}

