/*
 * DataLoad.C	.MEX file
 * load data from pre-existing file 
 * input: filename, output array name, channel name, duration
 *        (it takes all data if last two input arguements missing)
 * e.g. DataLoad('myfile', 'Out', 'IFO_DMRO', 3)
 */

#include <stdio.h>
#include "mex.h"
#include "mexversion.h"

#define CHANNUM     16

int getcharpos(char* name);

void mexFunction(
                 int nlhs,       mxArray *plhs[],
                 int nrhs, const mxArray *prhs[]
		 )
{
int   j, i, k, chidx=-1, Nchan, len, dur, res=1, multi, duration, datatype,
      trendtype=0, dataskip, bi, count, linelen[9+CHANNUM];
char  channel[80], chName[CHANNUM][80], filename[100], mexname[100], 
      tempst[200], output[80];
float tempf;
double  *outdata, *vout;
mxArray *array_ptr; 
FILE *fp;

  /* Check for proper number of arguments */
  
  if (nrhs < 2 )
    mexErrMsgTxt("you need to input a data file name and a output array name");

  if ( !mxIsChar(prhs[0]) ) {
    mexErrMsgTxt("The first argument should be the filename");
  }
  mxGetString (prhs[0], filename, 100);

  if ( !mxIsChar(prhs[1]) ) {
    mexErrMsgTxt("The second argument should be the name of the output array");
  }
  mxGetString (prhs[1], output, 100);

  /* open the file for information */

  fp = fopen(filename, "r");
  if ( fp == NULL ) {
    printf ( "Cann't open reading file: %s\n", filename );
    mexErrMsgTxt(" ");
  }
  /* Check for current version */
  fscanf ( fp, "%s", tempst ); 
  if (strncmp(tempst, "%dataVersion", 11))
    mexErrMsgTxt("Not a file generated by Data Mex.");
  sscanf ( tempst, "%%dataVersion%f-%d", &tempf, &bi );
  if ( tempf - VERSION > 0.01 || VERSION - tempf > 0.01)
    printf ( "WARNING: file %s was generated by Data Mex version %3.1f\n", filename, tempf );
  fscanf ( fp, "%s", tempst ); 
  sscanf ( tempst, "dataType=%d", &datatype );
  fscanf ( fp, "%s", tempst ); 
  sscanf ( tempst, "dataChnum=%d", &Nchan );
  if ( bi )
    if (mexEvalString(tempst)) 
      mexErrMsgTxt("Unexpeted error occured 1.\n");  
  fscanf ( fp, "%s", tempst );
  sscanf ( tempst, "dataDuration=%d", &dur );
  fscanf ( fp, "%s", tempst );
  if (datatype == 0) {
    sscanf ( tempst, "dataResolution=%d", &res );
    fscanf ( fp, "%s", tempst );
  }
  else {
    sscanf ( tempst, "dataTrend=%d", &trendtype );
    fscanf ( fp, "%s", tempst );
    sscanf ( tempst, "dataSkip=%d", &dataskip );
    res = 1;
  }
  fscanf ( fp, "%s", tempst ); 
  for ( j=0; j<Nchan-1; j++ ) {
    fscanf ( fp, "%s", chName[j] );
  } 
  fscanf ( fp, "%s", chName[Nchan-1] );
  fscanf ( fp, "%s", tempst );
  fscanf ( fp, "%s", tempst ); /* dataStart */
  fclose(fp);

  if ( bi == 0 ) {
    strcpy(mexname, filename);
    len = strlen(filename);
    for ( j=0; j<len; j++ ) {
      if ( filename[j] == '.' ) { 
	strncpy(mexname, filename, j);
	mexname[j] = '\0';
      } 
    }
    if (mexEvalString(mexname)) {
      mexPrintf("Error loading M-file '%s'", filename);
      mexErrMsgTxt("\n");
    }
  }
  else { /* load data from a binary file */
    fp = fopen(filename, "r");
    if ( fp == NULL ) {
      printf ( "Cann't open reading file: %s\n", filename );
      mexErrMsgTxt(" ");
    }
    for ( j=0; j<9+Nchan; j++ ) {
      fscanf ( fp, "%s", tempst );
      linelen[j] = strlen(tempst);
    } 
    fclose(fp);
    fp = fopen(filename, "r+b");
    if ( fp == NULL ) {
      printf ( "Cann't open reading file: %s\n", filename );
      mexErrMsgTxt(" ");
    }
    if ( datatype && trendtype == 0 )
      multi = 3;
    else
      multi = 1;
    vout = (double*)malloc(sizeof(double)*res*dur);
    if ( vout == NULL ) 
      mexErrMsgTxt("Error: no enough memory.\n");
    for ( j=0; j<9+Nchan; j++ ) 
      fread(tempst,sizeof(char),linelen[j]+1,fp);
    array_ptr = mxCreateDoubleMatrix(res*dur,Nchan*multi+1,mxREAL);
    outdata = mxGetPr(array_ptr);
    count = 0;
    for ( j=0; j<Nchan*multi+1; j++ ) { /* for each chan */
      fread(vout,sizeof(double),res*dur,fp);
      for ( i=0; i<res*dur; i++ ) {
	/*printf ( "j=%d vout[%d]=%f\n",j,i,vout[i] );*/
	*(outdata+count) = vout[i];
	count++;
      }
    }
    fclose(fp);
    free(vout);
    mxSetName(array_ptr, "dataArray");
    mexPutArray(array_ptr, "base");		 
    mxDestroyArray(array_ptr);
  }


  printf ( "DataLoad version %3.1f: file '%s'.\n", VERSION, filename );

  if ( nrhs == 2 ) { /* load all data */
    sprintf ( tempst, "%s=dataArray;", output );
    if( mexEvalString(tempst) )
      mexErrMsgTxt("unexpected error occured 2.\n");
    if ( datatype == 0 )
      printf ( "Done loading %d channel of raw data. The output array is %s.\n", Nchan, output );
    else if ( trendtype == 0 )
      printf ( "Done loading %d channel of trend min-mean-max data. The output array is %s.\n", Nchan, output );
    else
      printf ( "Done loading %d channel of trend %s data. The output array is %s.\n", Nchan, trendText[trendtype], output );
    mexEvalString("clear dataArray;"); 
    return;
  }

  if ( !mxIsChar(prhs[2]) ) { /* if more input submitted... */
    mexErrMsgTxt("The 3rd argument should be a channel name");
  }
  mxGetString (prhs[2], channel, 80);
  for ( j=0; j<Nchan; j++ ) {
    i = getcharpos(chName[j]);
    strncpy(tempst, chName[j]+1, i-1 );
    tempst[i-1] = '\0';
    strcpy(chName[j], tempst);
    if (strcmp(chName[j], channel) == 0 ) {
      chidx = j+2;
      break;
    }
  }
  if ( chidx < 0 ) {
    printf ( "Data for ch.%s is not found.\n", channel );
    mexErrMsgTxt(" ");
  }
  if (  nrhs > 3 && mxIsNumeric((prhs[3])) ) { 
    duration = (int)mxGetScalar(prhs[3]);
    if ( datatype == 60 )
      duration = duration/60;
    if ( duration > dur || duration < 0 ) {
      duration = dur;
      printf ( "WARNING: duration should be <= %d.\n", dur );
    }
  }
  else
    duration = dur;

  /* delete the rows beyond the duration */
  if ( duration < dur ) {
    
    for ( j=dur*res; j>duration*res; j-- ) {
      sprintf ( tempst, "dataArray(%d,:)=[];", j);
      if( mexEvalString(tempst) )
	mexErrMsgTxt("unexpected error occured 3.\n");
    }
  }

  if ( datatype == 0 || trendtype != 0 )
    sprintf ( tempst, "%s=[dataArray(:,1) dataArray(:,%d)];", output, chidx);
  else 
    sprintf ( tempst, "%s=[dataArray(:,1) dataArray(:,%d) dataArray(:,%d) dataArray(:,%d)];", output, 3*chidx-4,3*chidx-3,3*chidx-2);
  if( !mexEvalString(tempst) ) {
    printf ( "Done. the output array is %s.\n", output );
    if ( datatype == 0 )
      printf ( "Done loading raw data for %s. The output array is %s.\n", channel, output );
    else if  ( trendtype != 0 )
      printf ( "Done loading trend %s data for %s. The output array is %s.\n", trendText[trendtype], channel, output );
    else
      printf ( "Done loading trend min-mean-max data for %s. The output array is %s.\n", channel, output );
  }
  else
    mexErrMsgTxt("unexpected error occured 4.\n");
  mexEvalString("clear dataArray;"); 
  return;
}

int getcharpos(char* name)
{
int len, j;
  len = strlen(name);
  for ( j=1; j<len; j++ ) {
    if ( name[j] == '\'' ) 
      return j;
  }
  return 0;  
}

