/*
 * DataPlot.C	.MEX file
 * load data from pre-existing file and plot them
 * input: filename, channel name, duration
 *        (it takes all data if last two input arguements missing)
 * e.g. DataLoad('myfile', 'IFO_DMRO', 3)
 */

#include <stdio.h>
#include "mex.h"
#include "mexversion.h"

#define CHANNUM     16

int   fixchname(char* name);
void  fixtext(char* name);

void mexFunction(
                 int nlhs,       mxArray *plhs[],
                 int nrhs, const mxArray *prhs[]
		 )
{
int   j, i, chidx=-1, Nchan, len, dur, res=1, multi,duration, datatype, 
      trendtype=0, dataskip, bi, count, linelen[9+CHANNUM];
int   yy, mm, dd, hh, mn, ss;
char  channel[80], chName[CHANNUM][80],filename[100], mexname[100], 
      tempst[100], tempst1[200], timest[24], output[80];
float tempf;
double  *outdata, *vout;
mxArray *array_ptr; 
FILE *fp;

  strcpy(output, "dataOut"); 
  /* Check for proper number of arguments */
  
  if (nrhs < 1 )
    mexErrMsgTxt("you need to input a data file name");

  if ( !mxIsChar(prhs[0]) ) {
    mexErrMsgTxt("The first argument should be the filename");
  }
  mxGetString (prhs[0], filename, 100);

  /* open the file for information */

  fp = fopen(filename, "r");
  if ( fp == NULL ) {
    printf ( "Cann't open reading file: %s\n", filename );
    mexErrMsgTxt(" ");
  }
  /* Check for current version */
  fscanf ( fp, "%s", tempst ); 
  if (strncmp(tempst, "%dataVersion", 11))
    mexErrMsgTxt("Not a file generated by Data Mex.");
  sscanf ( tempst, "%%dataVersion%f-%d", &tempf, &bi );
  if ( tempf - VERSION > 0.01 || VERSION - tempf > 0.01)
    printf ( "WARNING: file %s was generated by Data Mex version %3.1f\n", filename, tempf );
  fscanf ( fp, "%s", tempst ); 
  sscanf ( tempst, "dataType=%d", &datatype );
  fscanf ( fp, "%s", tempst ); 
  sscanf ( tempst, "dataChnum=%d", &Nchan );
  if ( bi )
    if (mexEvalString(tempst)) 
      mexErrMsgTxt("Unexpeted error occured.\n");  
  fscanf ( fp, "%s", tempst );
  sscanf ( tempst, "dataDuration=%d", &dur );
  fscanf ( fp, "%s", tempst );
  if (datatype == 0) {
    sscanf ( tempst, "dataResolution=%d", &res );
    fscanf ( fp, "%s", tempst );
  }
  else {
    sscanf ( tempst, "dataTrend=%d", &trendtype );
    fscanf ( fp, "%s", tempst );
    sscanf ( tempst, "dataSkip=%d", &dataskip );
    res = 1;
  }
  fscanf ( fp, "%s", tempst ); 
  for ( j=0; j<Nchan-1; j++ ) {
    fscanf ( fp, "%s", chName[j] );
  } 
  fscanf ( fp, "%s", chName[Nchan-1] );
  fscanf ( fp, "%s", tempst );
  fscanf ( fp, "%s", tempst );
  sscanf ( tempst, "dataStart='%d-%d-%d-%d-%d-%d';", &yy,&mm,&dd,&hh,&mn,&ss );
  if ( yy < 10 )
    sprintf ( timest, "0%d-%d-%d-%d-%d-%d", yy,mm,dd,hh,mn,ss );
  else
    sprintf ( timest, "%d-%d-%d-%d-%d-%d", yy,mm,dd,hh,mn,ss );
  fclose(fp);

  if ( bi == 0 ) {
    strcpy(mexname, filename);
    len = strlen(filename);
    for ( j=0; j<len; j++ ) {
      if ( filename[j] == '.' ) { 
	strncpy(mexname, filename, j);
	mexname[j] = '\0';
      } 
    }
    if (mexEvalString(mexname)) {
      mexPrintf("Error loading M-file '%s'", filename);
      mexErrMsgTxt("\n");
    }
  }
  else { /* load data from a binary file */
    fp = fopen(filename, "r");
    if ( fp == NULL ) {
      printf ( "Cann't open reading file: %s\n", filename );
      mexErrMsgTxt(" ");
    }
    for ( j=0; j<9+Nchan; j++ ) {
      fscanf ( fp, "%s", tempst );
      linelen[j] = strlen(tempst);
    } 
    fclose(fp);
    fp = fopen(filename, "r+b");
    if ( fp == NULL ) {
      printf ( "Cann't open reading file: %s\n", filename );
      mexErrMsgTxt(" ");
    }
    if ( datatype && trendtype == 0 )
      multi = 3;
    else
      multi = 1;
    vout = (double*)malloc(sizeof(double)*res*dur);
    if ( vout == NULL ) 
      mexErrMsgTxt("Error: no enough memory.\n");
    for ( j=0; j<9+Nchan; j++ ) 
      fread(tempst,sizeof(char),linelen[j]+1,fp);
    array_ptr = mxCreateDoubleMatrix(res*dur,Nchan*multi+1,mxREAL);
    outdata = mxGetPr(array_ptr);
    count = 0;
    for ( j=0; j<Nchan*multi+1; j++ ) { /* for each chan */
      fread(vout,sizeof(double),res*dur,fp);
      for ( i=0; i<res*dur; i++ ) {
	*(outdata+count) = vout[i];
	count++;
      }
    }
    fclose(fp);
    free(vout);
    mxSetName(array_ptr, "dataArray");
    mexPutArray(array_ptr, "base");		 
    mxDestroyArray(array_ptr);
  }

  printf ( "DataPlot version %3.1f: file '%s'.\n", VERSION, filename );

  for ( j=0; j<Nchan; j++ ) { /* fix channel names */
    i = fixchname(chName[j]);
    strncpy(tempst, chName[j]+1, i-1 );
    tempst[i-1] = '\0';
    strcpy(chName[j], tempst);
  }

  if ( nrhs == 1 ) { /* plot all data - raw data only*/
    if ( datatype != 0 && trendtype == 0 )
      mexErrMsgTxt("Please specify a channel name.\n");
    sprintf ( tempst, "%s=dataArray;", output );
    if( mexEvalString(tempst) )
      mexErrMsgTxt("unexpected error occured.\n");
    printf ( "Data loaded. Ploting...\n" );
    sprintf ( tempst, "dX=%s(:,1);", output );
    if( mexEvalString(tempst) )
      mexErrMsgTxt("Unable plot the data. Unexpected error occured.\n");
    sprintf ( tempst1, "plot (" ); 
    for ( j=2; j<Nchan+1; j++ ) {
      sprintf ( tempst, "dX,%s(:,%d),",output,j ); 
      strcat(tempst1,tempst );
    }
    sprintf ( tempst, "dX,%s(:,%d));\n",output,Nchan+1 ); 
    strcat(tempst1,tempst );
    
    if( mexEvalString(tempst1) )
      mexErrMsgTxt("Unable plot the data. Unexpected error occured.\n");
    if ( datatype == 0 )
      sprintf ( tempst, "title('\\bf %d channels from %s ');", Nchan, timest);
    else
      sprintf ( tempst, "title('\\bf %d channels trend %s from %s ');", Nchan, trendText[trendtype], timest);
    mexEvalString(tempst);
    strcpy(tempst1, "legend(");
    for ( j=0; j<Nchan; j++ ) {
      fixtext(chName[j]);
      sprintf ( tempst, "'%s',", chName[j] );
      strcat(tempst1,tempst );
    }
    strcpy(tempst, "-1);");
    strcat(tempst1,tempst );
    mexEvalString(tempst1);
    if (datatype == 60) 
      mexEvalString("xlabel('time in minutes')");
    else
      mexEvalString("xlabel('time in seconds')");
    mexEvalString("ylabel('ADC count')");
    mexEvalString("grid on");
    
    mexEvalString("clear dataArray;"); 
    mexEvalString("clear dataOut;"); 
    mexEvalString("clear dX;"); 
    return;
  }

  /* if more input submitted... */

  if ( !mxIsChar(prhs[1]) ) {
    mexErrMsgTxt("The 2nd argument should be a channel name");
  }
  mxGetString (prhs[1], channel, 80);
  for ( j=0; j<Nchan; j++ ) {
    if (strcmp(chName[j], channel) == 0 ) {
      chidx = j+2;
      break;
    }
  }
  if ( chidx < 0 ) {
    printf ( "Data for ch.%s is not found.\n", channel );
    mexErrMsgTxt(" ");
  }
  if ( nrhs > 2 && mxIsNumeric((prhs[2])) ) { 
    duration = (int)mxGetScalar(prhs[2]);
    if ( datatype == 60 )
      duration = duration/60;
    if ( duration > dur || duration < 0 ) {
      duration = dur;
      printf ( "WARNING: duration should be <= %d.\n", dur );
    }
  }
  else
    duration = dur;

  /* delete the rows beyond the duration */
  if ( duration < dur ) {
    for ( j=dur*res; j>duration*res; j-- ) {
      sprintf ( tempst, "dataArray(%d,:)=[];", j);
      if( mexEvalString(tempst) )
	mexErrMsgTxt("unexpected error occured.\n");
    }
  }

  if ( datatype == 0 || trendtype != 0 )
    sprintf ( tempst, "%s=[dataArray(:,1) dataArray(:,%d)];", output, chidx);
  else 
    sprintf ( tempst, "%s=[dataArray(:,1) dataArray(:,%d) dataArray(:,%d) dataArray(:,%d)];", output, 3*chidx-4,3*chidx-3,3*chidx-2);
  if( !mexEvalString(tempst) )
    printf ( "Data loaded. Ploting...\n" );
  else
    mexErrMsgTxt("unexpected error occured.\n");

  fixtext(channel); /* ploting */
  if ( datatype == 0)
    sprintf ( tempst, "plot (%s(:,1), %s(:,2))\n title('\\bf %s   from %s ');",
          output, output, channel, timest );
  else{
    if ( trendtype == 0 ) {
      sprintf ( tempst, "plot (%s(:,1), %s(:,2),%s(:,1), %s(:,3),%s(:,1), %s(:,4))\n title('\\bf %s from %s ');",
	 output, output, output, output, output, output, channel, timest );
      strcat(tempst, "legend('min','mean','max',-1);");
    }
    else {
      sprintf ( tempst, "plot (%s(:,1), %s(:,2))\n title('\\bf %s trend %s from %s ');",
         output, output, channel, trendText[trendtype], timest );
    }
  }
  if( mexEvalString(tempst) )
    mexErrMsgTxt("Unable plot the data. Unexpected error occured.\n");
  if (datatype == 60) 
    mexEvalString("xlabel('time in minutes')");
  else
    mexEvalString("xlabel('time in seconds')");
  mexEvalString("ylabel('ADC count')");
  mexEvalString("grid on");
  		 
  mexEvalString("clear dataArray;"); 
  mexEvalString("clear dataOut;"); 
  mexEvalString("clear dX;"); 
  return;
}

/* grab the position of "'" in the line, used for getting chan names */
int fixchname(char* name)
{
int len, j;
  len = strlen(name);
  for ( j=1; j<len; j++ ) {
    if ( name[j] == '\'' ) 
      return j;
  }
  return 0;  
}

/* fix "_" in text, replace it with "\_" */
void  fixtext(char* name)
{
  
char tempname[100];
int len, i, j=0;
  strcpy(tempname, name);
  while ( name[j] != '\0' ) {
    if ( name[j] == '_' ) {
      strcpy(tempname, name);
      len = strlen(tempname);
      name[j] = '\\';
      for ( i=j+1; i<=len+1; i++ ) 
	name[i] = tempname[i-1];
      j++;
    }
    j++;
  }
  return;
}
